# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""Module for declaring the datatype.
"""
import base64

import numpy as np


class DataType:
    """
    Framework for supporting data types which
    are not JSON serializable by default.
    """

    _type = None
    object_key = None
    object_schema = None

    def can_encode(self, obj) -> bool:
        """
        Checks that the object can be
        encoded with this class. Default
        behaviour is to check that the
        object is an instance of _type.
        """
        if self._type is None:
            raise RuntimeError("_type not set for: {}".format(self))

        return isinstance(obj, self._type)

    def encode(self, obj) -> dict:
        """
        Encodes the object. Result should
        be JSON serializable. To be overridden
        by subclass.
        """
        raise NotImplementedError

    def can_decode(self, obj: dict) -> bool:
        """
        Checks that the object can be
        decoded with this class. Default
        behaviour is to check that the
        object_key exists in the object.
        """
        if self.object_key is None:
            raise RuntimeError("object_key not set for: {}".format(self))

        return self.object_key in obj

    def decode(self, obj: dict):
        """
        Decodes the object. To be overridden
        by subclass.
        """
        raise NotImplementedError


class NumpyArray(DataType):
    """Represent NumpyArray Model.
    """

    _type = np.ndarray
    object_key = "base64_encoded_array"

    def encode(self, obj):
        if obj.flags["C_CONTIGUOUS"]:
            obj_data = obj.data
        else:
            cont_obj = np.ascontiguousarray(obj)
            assert cont_obj.flags["C_CONTIGUOUS"]
            obj_data = cont_obj.data

        data_b64 = base64.b64encode(obj_data)
        return {
            self.object_key: data_b64.decode("utf-8"),
            "dtype": str(obj.dtype),
            "shape": list(obj.shape),
        }

    def decode(self, obj):
        return np.frombuffer(
            base64.b64decode(obj[self.object_key]), dtype=obj["dtype"]
        ).reshape(obj["shape"])


class NumpyComplexNumber(DataType):
    """Represent NumpyComplexNumber Model.
    """

    _type = np.complexfloating
    object_key = "base64_encoded_data"

    def encode(self, obj) -> dict:
        obj_data = obj.data
        data_b64 = base64.b64encode(obj_data)
        return {
            self.object_key: data_b64.decode("utf-8"),
            "dtype": str(obj.dtype),
            "shape": list(obj.shape),
        }

    def decode(self, obj: dict):
        cast_to = getattr(np, obj["dtype"])
        return cast_to(
            np.frombuffer(
                base64.b64decode(obj[self.object_key]), dtype=obj["dtype"]
            ).reshape(obj["shape"])
        )


class ComplexNumber(DataType):
    """Represents ComplexNumber Model.
    """

    _type = complex
    object_key = "encoded_complex"

    def encode(self, obj) -> dict:
        return {self.object_key: True, "real": obj.real, "imag": obj.imag}

    def decode(self, obj: dict):
        return complex(obj["real"], obj["imag"])
