# Copyright 2020 Q-CTRL Pty Ltd & Q-CTRL Inc. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#     https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""Module for serializing data
"""
import json
from typing import Dict, List, Union

import numpy as np

from .data_types import ComplexNumber, NumpyArray, NumpyComplexNumber


class DataTypeMixin:
    """
    Mixin class that defines the available
    data types in precedence order. To be
    used by the json encoder and decoder
    classes.
    """

    data_types = [NumpyArray(), NumpyComplexNumber(), ComplexNumber()]


class DataTypeDecoder(json.JSONDecoder, DataTypeMixin):
    """Represent DataTypeDecoder Model.
    """

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(
            self, object_hook=self.object_hook, *args, **kwargs
        )  # pylint:disable=method-hidden

    def object_hook(self, obj):
        """A function to check if any data type can be decoded.
        """
        for data_type in self.data_types:
            if data_type.can_decode(obj):
                return data_type.decode(obj)

        # otherwise decode as usual
        return obj


class DataTypeEncoder(json.JSONEncoder, DataTypeMixin):
    """Represent DataTypeEncoder Model.
    """

    def default(self, obj):  # pylint:disable=arguments-differ, method-hidden
        # check if any data type can be encoded
        for data_type in self.data_types:
            if data_type.can_encode(obj):
                return data_type.encode(obj)

        # otherwise let the base class default method raise the TypeError
        super().default(obj)
        return None


def encode(data: Union[np.ndarray, np.complex, complex, None]) -> dict:
    """

    Args:
      data: the data structure optionally containing complex handled
      by DataTypeEncoder.

    Returns:
      the encoded dict.

    """
    return json.loads(json.dumps(data, cls=DataTypeEncoder))


def decode(
    data: Union[List, Dict]
) -> Union[List, Dict, np.ndarray, np.complex, complex]:
    """

    Args:
      data: the data structure optionally containing objects handled
      by DataTypeDecoder.

    Returns:
        the decoded data structure

    """
    return json.loads(json.dumps(data), cls=DataTypeDecoder)
